/*
 * Utility to input touchscreen coordinates from user to shell scripts.
 * Additionally there is option to verifiy user input with tactile feedback.
 *
 * Authors:
 * Copyright 2012 by Jussi Ohenoja <juice@swagman.org>
 * Copyright 2012 by Maciej Jaros <egil@wp.pl>
 *
 * Ideas borrowed from evkey.c by Frantisek Dufka <dufkaf@seznam.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <poll.h>
#include <ctype.h>
#include <linux/input.h>

// if debug info needed
//#include "evtap_debug.c"

/**
	returns [s].[ms]
*/
double get_time_with_ms()
{
	struct timeval tim;
	gettimeofday(&tim, NULL);
	return (double) tim.tv_sec + (double) tim.tv_usec / 1000000.0;
}

#define TOUCH_BURST_TIMEOUT    -2
#define TOUCH_BURST_ERROR      -1
#define TOUCH_BURST_TOUCHSTART 1
#define TOUCH_BURST_TOUCHEND   2
#define TOUCH_BURST_TOUCHMOVE  3
/**
	Get touch burst from touch device.
	
	Every burst touch events ends with:
		EV_SYN:SYN_REPORT
	If it doesn't then the device was probably left in bad state
	(maybe because some program didn't wait for SYN_REPORT).

	@param tsdev touch device handle
	@param xcoord [out] coordinate read from input device
	@param ycoord [out] coordinate read from input device

	@return burstType:
		* lower then 0:
		** TOUCH_BURST_ERROR      - something unexpected happened
		** TOUCH_BURST_TIMEOUT    - haven't had any input for 100 ms (you might want to try again)
		* greater then 0:
		** TOUCH_BURST_TOUCHSTART - touch start indicator (x,y values are returned)
		** TOUCH_BURST_TOUCHEND   - touch end indicator (x,y values are empty; previously read x,y values are final)
		** TOUCH_BURST_TOUCHMOVE  - touch is in progress (x,y values are returned)
	
	Notes:
	EV_KEY:BTN_TOUCH = 1 - on touch start; occures before EV_ABS events
	EV_KEY:BTN_TOUCH = 0 - on touch end; occures after EV_ABS events
	EV_ABS:ABS_X = x - only occure on touch start or drag (move while touching)
	EV_ABS:ABS_Y = y - only occure on touch start or drag (move while touching)
	EV_ABS:ABS_MT_POSITION_X = x - occure when touching (when: start, hold, drag, end)
	EV_ABS:ABS_MT_POSITION_Y = y - occure when touching (when: start, hold, drag, end)
*/
int get_touch_burst(int tsdev, int *xcoord, int *ycoord)
{
	// prepare polling for timeout when device is not ready
	struct pollfd pfd;
	memset(&pfd, 0, sizeof(pfd));
	pfd.fd = tsdev;
	pfd.events = POLLIN;
	int timeout = 100;	// wait 100 ms for each burst
	
	int burstType = TOUCH_BURST_TOUCHMOVE;
	struct input_event ievent;
	int numCoordsRead = 0;
	int ret;
	while(1) {
		if(poll(&pfd, 1, timeout) < 1) { return TOUCH_BURST_TIMEOUT; }
		
		// read event data
		if((ret = read(tsdev ,&ievent ,sizeof(ievent))) < sizeof(ievent)) {
			printf("\nreceived only %d bytes on read", ret);
			return -2;
		}
		
		// debug: event info
		//debug_event_info (&ievent);

		switch(ievent.type) {
			case EV_SYN:
				if(ievent.code == SYN_REPORT) {
					if (numCoordsRead % 2 != 0)	 // should be 0 or 2
					{
						return TOUCH_BURST_ERROR;
					}
					return burstType;
				}
			break;
			case EV_KEY:
				if(ievent.code == BTN_TOUCH && ievent.value == 1) {
					burstType = TOUCH_BURST_TOUCHSTART;
				}
				else if(ievent.code == BTN_TOUCH && ievent.value == 0) {
					burstType = TOUCH_BURST_TOUCHEND;
				}
			break;
			case EV_ABS:
				if(ievent.code == ABS_MT_POSITION_X) {
					*xcoord = ievent.value;
					numCoordsRead++;
				}
				else if (ievent.code == ABS_MT_POSITION_Y) {
					*ycoord = ievent.value;
					numCoordsRead++;
				}
			break;
		}
	}
}

#define WAIT_MODE_END   0
#define WAIT_MODE_START 1
#define READ_MODE_FIRST 0
#define READ_MODE_LAST  1
/**
	Get touch coordinates from touch device.

	@param tsdev touch device handle
	@param timeout [in/out] (in ms)
		maximum delay allowed to wait for input;
		if 0 then waiting until specified event occures;
		upon succes will contain time spent waiting.
	@param xcoord [out] coordinate read from input device
	@param ycoord [out] coordinate read from input device
	@param waitMode 
		* WAIT_MODE_END   - read until touch end (or timeout)
		* WAIT_MODE_START - read until touch start (or timeout)
	@param readMode
		* READ_MODE_FIRST - get first x,y
		* READ_MODE_LAST  - get last x,y
	
	@return
		 0 upon succesful read
		-1 upon timeout (but only if coordinates were not read)
		-2 upon error (should not happen if device is correct)
*/
int get_touchtap(int tsdev, int *timeout, int *xcoord, int *ycoord, int waitMode, int readMode)
{
	// prepare polling
	struct pollfd pfd;
	memset(&pfd, 0, sizeof(pfd));
	pfd.fd = tsdev;
	pfd.events = POLLIN;
	
	// read events from input
	double startTime = get_time_with_ms();
	int origTimeout = *timeout;
	int burstType = 0;
	int burstX, burstY;
	int numCoordReads = 0;
	burstX = burstY = 0;
	while(1) {
		burstType = get_touch_burst (tsdev, &burstX, &burstY);
		
		//fprintf(stderr, "\n[%d] = (%d, %d)", burstType, burstX, burstY);
		
		// burst OK?
		if (burstType > 0)
		{
			// touch-start mode
			if(waitMode == WAIT_MODE_START) {
				if (burstType != TOUCH_BURST_TOUCHSTART) {
					continue;
				}
				*xcoord = burstX;
				*ycoord = burstY;
				return 0;
			}
			// touch-end mode
			else {
				if (burstType == TOUCH_BURST_TOUCHEND) {
					return 0;
				}
				// read first x,y mode
				if (readMode == READ_MODE_FIRST) {
					if(numCoordReads < 1) {
						numCoordReads++;
						*xcoord = burstX;
						*ycoord = burstY;
					}
				}
				// read last x,y mode
				else {
					numCoordReads++;
					*xcoord = burstX;
					*ycoord = burstY;
				}
			}
		}
		
		// time check
		double curTime = get_time_with_ms();
		*timeout = (int)(1000.0 * (curTime - startTime));
		// timeout if it was given and time gone
		if(origTimeout) {
			if(*timeout >= origTimeout) {
				if(numCoordReads >= 1) {
					return 0;	// OK - we got what we want
				}
				else {
					return -1;	// unable to read coords in given timeout
				}
			}
		}
	}
}


int send_vibra(int videv)
{
	struct input_event ievent;
	struct timespec ts;
	int ret, i;

	if(videv == 0) { return 0; }

	i = 20;
	while(i-->0) {
		ts.tv_sec = 0;
		ts.tv_nsec = 1000000;
		nanosleep(&ts, &ts);
		ievent.type = EV_FF;
		ievent.code = 20;
		ievent.value = 1;
		if((ret = write(videv, &ievent ,sizeof(ievent))) < sizeof(ievent)) {
			printf("\nsent only %d bytes on write", ret);
			return -1;
		}
	}
	
	return 0;
}


char *cleanup_line(char *line)
{
	int i=0;
	char *cp=NULL;

	while(i < strlen(line)) {
		if(!isblank(line[i])) {
			cp = line + i;
			break;
		}
		i++;
	}

	if(strlen(cp) == 0) { return NULL; }
	if(cp[0] == '#') { return NULL; }

	return cp;
}


struct mapline {
	int x1;
	int y1;
	int x2;
	int y2;
	int val;
	char c;
};


int get_mapline(FILE *mapfile, struct mapline *ml)
{
	char buf[200], *cp, *tok=NULL;

	memset(ml, 0, sizeof(struct mapline));
	if(fgets(buf, sizeof(buf), mapfile) != NULL) {
		if((cp = cleanup_line(buf)) == NULL) {
			return 0;  // comment line is ignored
		} else {
			tok = strtok(cp, " ");
			if(tok == NULL) { return 0; }
			ml->x1 = strtoul(tok, NULL, 0);
			tok = strtok(NULL, " ");
			if(tok == NULL) { return 0; }
			ml->y1 = strtoul(tok, NULL, 0);
			tok = strtok(NULL, " ");
			if(tok == NULL) { return 0; }
			ml->x2 = strtoul(tok, NULL, 0);
			tok = strtok(NULL, " ");
			if(tok == NULL) { return 0; }
			ml->y2 = strtoul(tok, NULL, 0);
			tok = strtok(NULL, " ");
			if(tok == NULL) { return 0; }
			ml->val = strtoul(tok, NULL, 0);
			tok = strtok(NULL, " ");
			if(tok == NULL) { return 0; }
			//ml->c = strtoul(tok, NULL, 0);
			ml->c = *tok;
			return 1;
		}
	}
	return -1;
}


int output_mapped_char(FILE *mapfile, int xcoord, int ycoord, int asciiFlag)
{
	struct mapline ml;
	int ret;

	while((ret = get_mapline(mapfile, &ml)) >= 0) {
		if(ret > 0) {
			if( (ml.x1 < xcoord) && (xcoord <= ml.x2) &&
				(ml.y1 < ycoord) && (ycoord <= ml.y2) ) {
				if (asciiFlag)
					printf("%d", ml.c);
				else
					printf("%c", ml.c);
				return ml.val;
			}
		}
	}
	
	return -1;
}


int get_mapped_value(FILE *mapfile, int xcoord, int ycoord)
{
	struct mapline ml;
	int ret;

	while((ret = get_mapline(mapfile, &ml)) >= 0) {
		if(ret > 0) {
			if( (ml.x1 < xcoord) && (xcoord <= ml.x2) &&
				(ml.y1 < ycoord) && (ycoord <= ml.y2) ) {
				return ml.val;
			}
		}
	}
	return -1;
}

int main(int argc, char **argv)
{
	int opt, tsdev, videv, xflag, yflag, cflag, asciiFlag, xcoord, ycoord, timeout, ret;
	static int waitMode = WAIT_MODE_START;
	static int readMode = READ_MODE_LAST;
	static int printTimeSpent = 0;
	FILE *mapfile=NULL;

	tsdev = videv = xflag = yflag = cflag = asciiFlag = xcoord = ycoord = timeout = 0;
	static struct option long_options[] =
	{
		// directly set flag
		{"wait-for-start", no_argument, &waitMode, WAIT_MODE_START},
		{"wait-for-end"  , no_argument, &waitMode, WAIT_MODE_END},
		{"read-first", no_argument, &readMode, READ_MODE_FIRST},
		{"read-last" , no_argument, &readMode, READ_MODE_LAST},
		{"print-time-spent" , no_argument, &printTimeSpent, 1},
		// without argument
		{"print-x"         , no_argument,     0, 'x'},
		{"print-y"         , no_argument,     0, 'y'},
		{"print-character" , no_argument,     0, 'c'},
		{"print-ascii-code", no_argument,     0, 'a'},
		{"help"            , no_argument,     0, 'h'},
		// with argument
		{"touch-dev" , required_argument, 0, 't'},
		{"vibra-dev" , required_argument, 0, 'v'},
		{"map-file"  , required_argument, 0, 'm'},
		{"wait-delay", required_argument, 0, 'd'},
		// end
		{0, 0, 0, 0}
	};
	
	int option_index = 0;
	while ( (opt = getopt_long (argc, argv, "t:v:m:d:M:xycha", long_options, &option_index)) != -1 ) {
		switch (opt) {
		case 't':
			if((tsdev = open(optarg, O_RDONLY)) < 0) {
				perror("cannot open touch input device");
				return -2;
			}
			break;
		case 'v':
			if((videv = open(optarg, O_RDWR)) < 0) {
				perror("cannot open vibration output device");
				return -2;
			}
			break;
		case 'm':
			if((mapfile = fopen(optarg, "r")) == NULL) {
				perror("cannot open mapping table");
				return -2;
			}
			break;
		case 'd':
			timeout = atoi(optarg);
			break;
		case 'x':
			xflag = 1;
			break;
		case 'y':
			yflag = 1;
			break;
		case 'c':
			cflag = 1;
			break;
		case 'a':
			asciiFlag = 1;
			cflag = 1;
			break;
		case 'h':
		case '?':
			printf("Version: 2012-11-03-nux-fork\n\n");
			printf("Usage:\n");
			printf("   evtap -t <touch-device-path> [OTHER-OPTIONS]\n\n");
			printf("   This utility reads coordinates of a screen tap, with optional vibration\n");
			printf("   serving as a tactile feedback to the user.\n");
			printf("   If delay is specified and no input is given then utility exits printing\n");
			printf("   \"timeout\" to stdout.\n\n");
			printf("Options:\n");
			printf("   --help -h, optional\n");
			printf("     display this help info and exit.\n");
			printf("\n");
			printf("   --touch-dev=<device-path> -t <device-path> (mandatory)\n");
			printf("     touch input device.\n");
			printf("   --vibra-dev=<device-path> -v <device-path> (optional)\n");
			printf("     vibra output device. If option is not present there is no tactile feedback.\n");
			printf("\n");
			printf("   --map-file=<file-path> -m <file-path> (optional)\n");
			printf("     Mapping file name. Without mapping file <-c> cannot be used.\n");
			printf("     If used map file must contain a mapping table where each line contains\n");
			printf("     following characters (seprated by spaces):\n");
			printf("     <x1> <y1> <x2> <y2> <numeric-value> <character>\n");
			printf("\n");
			printf("   --wait-for-start (default)\n");
			printf("     wait until touch-start occures (or delay ends)\n");
			printf("   --wait-for-end (optional)\n");
			printf("     wait until touch-end occures (or delay ends)\n");
			printf("   --read-first (optional)\n");
			printf("     return first value\n");
			printf("   --read-last (default)\n");
			printf("     return last value \n");
			printf("   --wait-delay=<ms> -d <ms> (optional)\n");
			printf("     delay in milliseconds (if 0 then waiting until event occures).\n");
			printf("\n");
			printf("   --print-x -x\n");
			printf("     Printout x coordinate.\n");
			printf("   --print-y -y\n");
			printf("     Printout y coordinate. If both x and y are given then coordinates are \n");
			printf("     separated witha a colon ':'\n");
			printf("   --print-ascii-code -a\n");
			printf("     Same as --print-character except that ascii value of the character is\n");
			printf("     returned rather then the character.\n");
			printf("   --print-character -c\n");
			printf("     Printout the character found in the map file.\n");
			printf("     The exit code is the corespoding numeric-value.\n\n");
			printf("   --print-time-spent\n");
			printf("     Printout time spent waiting for input [ms].\n");
			printf("     Works with any of above and is separated with semicolon ';' from them.\n");
			printf("Exit codes:\n");
			printf("   On success:\n");
			printf("     In x or y output modes returns 0.\n");
			printf("     In character and ascii-code output modes returns a numeric value\n");
			printf("     from the map file.\n");
			printf("   On timeout the utility returns -1.\n");
			printf("   On error cases the utility returns -2.\n");
			return 0;
		}
	}

	if(tsdev == 0) {
		printf("touch input device is mandatory parameter\n");
		return -2;
	}

	if(xflag == 0 && yflag == 0 && cflag == 0) {
		printf("output mode must be specified\n");
		return -2;
	}

	if((mapfile == NULL) && (cflag == 1)) {
		printf("there is no mapping table\n");
		return -2;
	}

	ret = get_touchtap(tsdev, &timeout, &xcoord, &ycoord, waitMode, readMode);
	if(ret == -2) { return -2; }
	if(ret == -1) { printf("timeout\n"); return -1; }
	if(send_vibra(videv) != 0) { return -2; }

	/* cflag gets predecence over other flags, we output only this and exit */
	if(cflag) {
		ret = output_mapped_char(mapfile, xcoord, ycoord, asciiFlag);
		if(printTimeSpent) { printf(";%d", timeout); }
		return ret;
	}

	if(xflag) { printf("%d", xcoord); }
	if(xflag && yflag) { printf(":"); }
	if(yflag) { printf("%d", ycoord); }
	if(printTimeSpent) { printf(";%d", timeout); }

	printf("\n");

	return 0;
}
